/* Copyright (C) 2003-2005 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_common.h"
#include "dps_log.h"
#include "dps_uniconv.h"
#include "dps_vars.h"
#include "dps_textlist.h"
#include "dps_utils.h"
#include "dps_parsexml.h"
#include "dps_hrefs.h"
#include "dps_sgml.h"

#ifdef WITH_XML

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <expat.h>

#ifndef XML_STATUS_ERROR
#define XML_STATUS_ERROR 0
#endif

/*
#define DEBUG_XML 1
*/

typedef struct {
  DPS_AGENT *Indexer;
  DPS_DOCUMENT *Doc;
  int body_sec;
  char *sec;
  char *secpath;
  size_t pathlen, curlen;
} XML_PARSER_DATA;


static void startElement(void *userData, const char *name, const char **attr) {
       XML_PARSER_DATA *D = userData;
       DPS_AGENT *Indexer = D->Indexer;
       DPS_DOCUMENT *Doc = D->Doc;
       size_t i;

       DPS_FREE(D->sec);
       D->sec = (char*)DpsStrdup(name);

       if (D->curlen + (i = strlen(name)) + 1 > D->pathlen) {
	 D->pathlen += 256 + i;
	 D->secpath = (char*)DpsRealloc(D->secpath, D->pathlen);
	 if (D->curlen == 0) D->secpath[0] = '\0';
       }
       D->curlen += i + ((D->secpath[0]) ? 1 : 0);
       sprintf(DPS_STREND(D->secpath), "%s%s", (D->secpath[0]) ? "|" : "", name);

#ifdef DEBUG_XML
       fprintf(stderr, "START sec:%s  secpath:%s  curlen:%d  pathlen:%d  i:%d\n", D->sec, D->secpath, D->curlen, D->pathlen, i);
#endif


       for (i = 0; attr[i]; i += 2) {
	 if (!strcasecmp(attr[i], "href")) {
		DPS_HREF	Href;
	   
/*		DpsSGMLUnescape(attr[i+1]);  we do this later */
		DpsHrefInit(&Href);
		Href.referrer = DpsVarListFindInt(&Doc->Sections, "Referrer-ID", 0);
		Href.hops = 1 + DpsVarListFindInt(&Doc->Sections, "Hops", 0);
		Href.site_id = DpsVarListFindInt(&Doc->Sections, "Site_id", 0);
		Href.url = attr[i+1];
		Href.method = DPS_METHOD_GET;
		DpsHrefListAdd(Indexer, &Doc->Hrefs, &Href);
	 }
       }


}

static void endElement(void *userData, const char *name) {
       XML_PARSER_DATA *D = userData;
       char *p = strrchr(D->secpath, '|');
       
       DPS_FREE(D->sec);
       if (p == NULL) {
	 D->secpath[0] = '\0';
	 D->curlen = 0;
       } else {
	 D->curlen -= strlen(p);
	 *p = '\0';
	 p++;
	 D->sec = (char*)DpsStrdup(p);
       }
#ifdef DEBUG_XML
       fprintf(stderr, "END   sec:%s  secpath:%s  curlen:%d  pathlen:%d\n", D->sec ? D->sec : "<NULL>", D->secpath, D->curlen, D->pathlen);
#endif
}


static void Decl(void *userData, const XML_Char *version, const XML_Char *encoding, int standalone) {
       XML_PARSER_DATA *D = userData;
       DPS_DOCUMENT *Doc = D->Doc;

       if (encoding != NULL) {
	 DpsVarListReplaceStr(&Doc->Sections, "Meta-Charset", DpsCharsetCanonicalName(encoding));
       }

}

static void Text(void *userData, const XML_Char *s, int len) {
        XML_PARSER_DATA *D = userData;
        DPS_DOCUMENT *Doc = D->Doc;
	DPS_AGENT *Indexer = D->Indexer;
        DPS_TEXTITEM	Item;
	DPS_VAR		*Sec;

	bzero((void*)&Item, sizeof(Item));
	Item.str = DpsStrndup(s, (size_t)len);
	if((D->sec != NULL) &&  (Sec = DpsVarListFind(&Indexer->Conf->HrefSections, D->sec))) {
	  Item.section = Sec->section;
	  Item.section_name = D->sec;
	} else if((D->sec != NULL) &&  (Sec = DpsVarListFind(&Doc->Sections, D->sec))){
	  Item.section = Sec->section;
	  Item.section_name = D->sec;
	} else {
	  Item.section = D->body_sec;
	  Item.section_name = "body";
	}
	DpsTextListAdd(&Doc->TextList, &Item);
	DpsFree(Item.str);
}

static int EncHandler(void *encodingHandlerData, const XML_Char *name, XML_Encoding *info) {
  DPS_AGENT *indexer = encodingHandlerData;
  DPS_CHARSET	*cs;
  size_t i;

  if (!(cs = DpsGetCharSet(name))) {
    return 0;
  }
  if (cs->tab_to_uni == NULL) return 0; /* FIXME: rewrite this for multibytes encodings */

  info->convert = NULL;
  info->release = NULL;
  info->data = NULL;
  for(i = 0; i < 256; i++) info->map[i] = cs->tab_to_uni[i];

  return 1;
}

int DpsXMLParse(DPS_AGENT *Indexer, DPS_DOCUMENT *Doc){
        int res = DPS_OK;
        XML_PARSER_DATA Data;
        XML_Parser parser = XML_ParserCreate(NULL);
	DPS_VAR		*BSec=DpsVarListFind(&Doc->Sections,"body");
	int		body_sec  = BSec ? BSec->section : 0;

	bzero(&Data, sizeof(Data));
	Data.Indexer = Indexer;
	Data.Doc = Doc;
	Data.body_sec = body_sec;

       XML_SetUserData(parser, &Data);
       XML_SetElementHandler(parser, startElement, endElement);
       XML_SetXmlDeclHandler(parser, Decl);
       XML_SetCharacterDataHandler(parser, Text);
       XML_SetUnknownEncodingHandler(parser, EncHandler, Indexer);

       if (XML_Parse(parser, Doc->Buf.content, (int)strlen(Doc->Buf.content), 1) == XML_STATUS_ERROR) {
	 DpsLog(Indexer, DPS_LOG_ERROR,
              "XML parsing error: %s at line %d\n",
              XML_ErrorString(XML_GetErrorCode(parser)),
              XML_GetCurrentLineNumber(parser));
	 res = DPS_ERROR;
       }
       XML_ParserFree(parser);

       DPS_FREE(Data.sec);
       DPS_FREE(Data.secpath);

       return res;
}

#else
char a[]="";
#endif
