/* Copyright (C) 2003-2005 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/
#include "dps_common.h"
#include "dps_utils.h"
#include "dps_stopwords.h"
#include "dps_unicode.h"
#include "dps_unidata.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>


DPS_STOPWORD * DpsStopListFind(DPS_STOPLIST *List, const dpsunicode_t *word, const char *lang) {
	int low  = 0;
	int high = List->nstopwords - 1;

	if(!List->StopWord) return NULL;
	while (low <= high) {
		int middle = (low + high) / 2;
		int match = DpsUniStrCmp(List->StopWord[middle].uword, word);
		if (match < 0) { low = middle + 1;
		} else if (match > 0) { high = middle - 1;
		} else {
		  if (lang==NULL || *lang=='\0' || !strncasecmp(List->StopWord[middle].lang, lang, strlen(List->StopWord[middle].lang)))
		    return &List->StopWord[middle];
		  return NULL;
		}
	}
	return NULL;
}

static int cmpstop(const void *s1, const void *s2){
	return(DpsUniStrCmp(((const DPS_STOPWORD*)s1)->uword, ((const DPS_STOPWORD*)s2)->uword));
}

void DpsStopListSort(DPS_STOPLIST *List){
	/* Sort stoplist to run binary search later */
	if (List->nstopwords > 1) DpsSort(List->StopWord, List->nstopwords, sizeof(DPS_STOPWORD), cmpstop);
}

int DpsStopListAdd(DPS_STOPLIST *List, DPS_STOPWORD * stopword) {
	size_t j;

	/* If the word is already in list     */
	/* We will not add it again           */
	/* But mark it as "international word"*/
	/* i.e. the word without language     */
	/* It will allow to avoid troubles    */
	/* with language guesser              */
	for(j=0;j<List->nstopwords;j++){
		if(!DpsUniStrCmp(List->StopWord[j].uword, stopword->uword)){
			DPS_FREE(List->StopWord[j].lang);
			List->StopWord[j].lang = (char*)DpsStrdup("");
			return 0;
		}
	}

	List->StopWord=(DPS_STOPWORD *)DpsRealloc(List->StopWord,(List->nstopwords+1)*sizeof(DPS_STOPWORD));
	if (List->StopWord == NULL) {
	  List->nstopwords = 0;
	  return 0;
	}
	List->StopWord[List->nstopwords].word = NULL;
	List->StopWord[List->nstopwords].uword = DpsUniDup(stopword->uword);
	List->StopWord[List->nstopwords].lang = (char*)DpsStrdup(stopword->lang?stopword->lang:"");
	List->StopWord[List->nstopwords].len = 0;
	List->StopWord[List->nstopwords].ulen = DpsUniLen(stopword->uword);
	List->nstopwords++;

	return(1);
}

void DpsStopListFree(DPS_STOPLIST *List){
	size_t i;
	for(i=0;i<List->nstopwords;i++){
		DPS_FREE(List->StopWord[i].uword);
		DPS_FREE(List->StopWord[i].word);
		DPS_FREE(List->StopWord[i].lang);
	}
	DPS_FREE(List->StopWord);
	List->nstopwords=0;
}


__C_LINK int __DPSCALL DpsStopListLoad(DPS_ENV * Conf, const char *fname) {
	char str[1024];
	char * lasttok;
	FILE * stopfile;
	DPS_STOPWORD stopword;
	DPS_CHARSET *cs = NULL, *uni_cs = DpsGetCharSet("sys-int");
	DPS_CONV cnv;
	char * charset=NULL;
	dpsunicode_t *uwrd, *nfc;

	if (!(stopfile=fopen(fname,"r")))
	{
		sprintf(Conf->errstr,"Can't open stopwords file '%s' (%s)", fname, strerror(errno));
		return DPS_ERROR;
	}
	if ((uwrd = (dpsunicode_t*)DpsMalloc(sizeof(dpsunicode_t) * (Conf->WordParam.max_word_len + 1))) == NULL) return DPS_ERROR;

	bzero((void*)&stopword, sizeof(stopword));

	while(fgets(str,sizeof(str),stopfile)){
		if(!str[0])continue;
		if(str[0]=='#')continue;
		
		if(!strncmp(str,"Charset:",8)){
			DPS_FREE(charset);
			charset = dps_strtok_r(str + 8, " \t\n\r", &lasttok);
			if(charset){
				charset = (char*)DpsStrdup(charset);
			}
		}else
		if(!strncmp(str,"Language:",9)){
			DPS_FREE(stopword.lang);
			stopword.lang = dps_strtok_r(str + 9, " \t\n\r", &lasttok);
			if(stopword.lang)stopword.lang = (char*)DpsStrdup(stopword.lang);
		}else
		if((stopword.word = dps_strtok_r(str, "\t\n\r", &lasttok))) {
			
			if(!cs){
				if(!charset){
					sprintf(Conf->errstr,"No charset definition in stopwords file '%s'", fname);
					DPS_FREE(stopword.lang);
					DPS_FREE(uwrd);
					return DPS_ERROR;
				}else{
					cs=DpsGetCharSet(charset);
					if(!cs){
						sprintf(Conf->errstr,"Unknown charset '%s' in stopwords file '%s'", charset,fname);
						DPS_FREE(stopword.lang);
						DPS_FREE(charset);
						DPS_FREE(uwrd);
						return DPS_ERROR;
					}
					DpsConvInit(&cnv, cs, uni_cs, DPS_RECODE_HTML);
				}
			}
			
			DpsConv(&cnv, (char*)uwrd, sizeof(dpsunicode_t) * Conf->WordParam.max_word_len, 
				stopword.word, strlen(stopword.word) + 1);
			uwrd[Conf->WordParam.max_word_len] = '\0';
			nfc = DpsUniNormalizeNFC(NULL, uwrd);
			stopword.uword = nfc;
			DpsStopListAdd(&Conf->StopWords, &stopword);
			DPS_FREE(nfc);

		}
	}
	fclose(stopfile);
	DpsStopListSort(&Conf->StopWords);
	DPS_FREE(stopword.lang);
	DPS_FREE(charset);
	DPS_FREE(uwrd);
	return DPS_OK;
}
